module RoomTool

using ..Ahorn, Maple
using Gtk

displayName = "Room"
group = "Rooms"

log(s) = Ahorn.debug.log(s, "TEMP_DEBUG")

# Globals
room_original_position = nothing
room_original_size = nothing
room_original_tiles = (nothing, nothing, nothing)
original_mouse_coords = nothing  # Where drag started
relevantRoom = nothing
subtool = "Move"
wasMouseHeld = false

# Moving
draggingRoom = false
onHorizontal = false
onVertical = false

# Resizing
resizingRoom = false
isLeftEdge = false
isTopEdge = false
isRightEdge = false
isBottomEdge = false

# Minimum room size this tool allows (1 screen)
roomMinWidth = 40 * 8
roomMinHeight = 23 * 8
roomMinSize = (roomMinWidth, roomMinHeight)

function drawNothing(layer::Ahorn.Layer, room::Ahorn.DrawableRoom, camera::Ahorn.Camera)
end

function toolSelected(subTools::Ahorn.ListContainer, layers::Ahorn.ListContainer, materials::Ahorn.ListContainer)
    global relevantRoom = Ahorn.loadedState.room
    Ahorn.updateLayerList!(["all"], 1)
    global subtool = get(Ahorn.persistence, "room_tool_mode", "Move")
    Ahorn.updateTreeView!(subTools, ["Move", "Resize", "Filler Resize"], row -> row[1] == subtool)
    # Don't set any material, they are empty by default
    # This tool doesn't draw anything on the tools layer
    Ahorn.redrawingFuncs["tools"] = drawNothing
    resetCursor()
end

function subToolSelected(list::Ahorn.ListContainer, selected::String)
    Ahorn.persistence["room_tool_mode"] = selected
    global subtool = selected
end

# Make sure size and position are correct after Undo or Redo
# Does not work if a different tool is selected before undoing
# TODO submit Ahorn patch
afterUndo(map::Maple.Map) = updateRoomList()
afterRedo(map::Maple.Map) = updateRoomList()

#TODO is replacement necessary/desirable?
updateRoomList() = Ahorn.updateTreeView!(Ahorn.roomList, Ahorn.getTreeData(Ahorn.loadedState.map), row -> row[1] == Ahorn.loadedState.room.name, updateByReplacement=true)

function roomChanged(room::Ahorn.DrawableRoom)
    roomResizeEnd()
    roomDragEnd()
    resetCursor()
    global relevantRoom = room
end

function cleanup()
    roomResizeEnd()
    roomDragEnd()
    resetCursor()
    global relevantRoom = Ahorn.loadedState.room
end

resetCursor() = Ahorn.setWindowCursor!(Ahorn.window, "default")
function updateCursor()
    # If we're dragging or resizing, keep the same cursor
    if subtool == "Move" && !draggingRoom
        cursor = "default"
        if onHorizontal && onVertical
            cursor = "move"
        end
        Ahorn.setWindowCursor!(Ahorn.window, cursor)
    elseif (subtool == "Resize" || subtool == "Filler Resize") && !resizingRoom
        cursor = "default"
        if (isLeftEdge && isTopEdge) || (isRightEdge && isBottomEdge)
            cursor = "nwse-resize"
        elseif (isRightEdge && isTopEdge) || (isLeftEdge && isBottomEdge)
            cursor = "nesw-resize"
        elseif isTopEdge || isBottomEdge
            cursor = "ns-resize"
        elseif isLeftEdge || isRightEdge
            cursor = "ew-resize"
        end
        Ahorn.setWindowCursor!(Ahorn.window, cursor)
    end
end

# Hook leave-notify-event ourselves to update cursor properly
# happens when mouse moves off canvas to a menu like tools/rooms
# In this case mouseHeld was already set to false (?)
# and we no longer receive mouseMotion calls
signal_connect(Ahorn.canvas, "leave-notify-event") do canvas, event
    # only act on the tools we defined
    if Ahorn.currentTool !== nothing
        toolname = Ahorn.getToolName(Ahorn.currentTool)
        if toolname == "Room"
            resetCursor()
        end
    end
    return false
end

# Hook mouse motion to get more control
function mouseMotion(event::Ahorn.eventMouse, camera::Ahorn.Camera)
    global wasMouseHeld
    abs_coords = Ahorn.getMapCoordinatesAbs(camera, event.x, event.y)

    # Continue a resize or drag
    if Ahorn.mouseButtonHeld(0x1)
        if resizingRoom
            roomResizeUpdate(abs_coords)
            wasMouseHeld = true
            return
        elseif draggingRoom
            roomDragUpdate(abs_coords)
            wasMouseHeld = true
            return
        end
    end

    setRoomHovering(abs_coords)

    if Ahorn.mouseButtonHeld(0x1)
        if wasMouseHeld
            #TODO
            # log("unexpected mouse held when not dragging")
            return
        end
        # New click - start dragging
        if subtool=="Move"
            if onHorizontal && onVertical
                roomDragBegin(abs_coords)
            end
        elseif subtool=="Resize" || subtool=="Filler Resize"
            if isLeftEdge || isRightEdge || isBottomEdge || isTopEdge
                roomResizeBegin(abs_coords)
            end
        end
        wasMouseHeld = true
    else
        roomDragEnd()
        roomResizeEnd()
        wasMouseHeld = false
    end
    updateCursor()
end

function setRoomHovering(abs_coords::Tuple{Int,Int})
    # Compute which parts of the room the mouse is hovering over, if any
    ax, ay = abs_coords
    leftEdge, topEdge = Int.(relevantRoom.position)
    rightEdge, bottomEdge = Int.(relevantRoom.position) .+ Int.(relevantRoom.size)
    if subtool == "Move"
        global onHorizontal = leftEdge <= ax < rightEdge
        global onVertical = topEdge <= ay < bottomEdge

    elseif subtool == "Resize" || subtool == "Filler Resize"
        inD = 16  # distance inside room edge: 2 tiles
        outD = 16  # distance outside room edge: 2 tiles
        # TODO outD > 0 will cause extra history entry on click
        # at edge of adjacent room to select it, even if no resize occurs
        # TODO implement inD/outD based on camera zoom level???

        onHorizontal = leftEdge - outD <= ax < rightEdge + outD
        onVertical = topEdge - outD <= ay < bottomEdge + outD

        global isLeftEdge = onVertical && leftEdge-outD <= ax < leftEdge+inD
        global isRightEdge = onVertical && rightEdge-inD <= ax < rightEdge+outD
        global isTopEdge = onHorizontal && topEdge-outD <= ay < topEdge+inD
        global isBottomEdge = onHorizontal && bottomEdge-inD <= ay < bottomEdge+outD
        # Don't allow two opposite edges at the same time
        if isTopEdge && isBottomEdge
            isTopEdge = false
        end
        if isLeftEdge && isRightEdge
            isLeftEdge = false
        end
    end
end


#= Move Room =#

function roomDragBegin(mouse_coords::Tuple{Int,Int})
    Ahorn.History.addSnapshot!(Ahorn.History.RoomSnapshot("Room Movement", relevantRoom))
    global draggingRoom = true
    global room_original_position = relevantRoom.position
    global original_mouse_coords = mouse_coords
end

function roomDragUpdate(new_mouse_coords::Tuple{Int,Int})
    mouse_change = new_mouse_coords .- original_mouse_coords
    tiles_change = floor.(Int, mouse_change ./ 8)
    room_new_position = room_original_position .+ (tiles_change .* 8)
    if relevantRoom.position == room_new_position
        # room is already there - can skip updating
        return
    end
    # move the room
    relevantRoom.position = room_new_position
    updateRoomList()
    # Update the map to show the room has moved
    Ahorn.draw(Ahorn.canvas)
end

function roomDragEnd()
    if draggingRoom
        # The room is already in place, we just have to stop moving it.
        global draggingRoom = false
        # reset other things just in case
        global room_original_position = nothing
        global original_mouse_coords = nothing
    end
end


#= Resize Room =#

function roomResizeBegin(abs_coords::Tuple{Int,Int})
    Ahorn.History.addSnapshot!(Ahorn.History.RoomSnapshot("Room Resize", relevantRoom))
    global resizingRoom = true
    global room_original_position = Int.(relevantRoom.position)
    global room_original_size = Int.(relevantRoom.size)
    global room_original_tiles = (copy(relevantRoom.fgTiles.data),
                                  copy(relevantRoom.bgTiles.data),
                                  copy(relevantRoom.objTiles.data))
    global original_mouse_coords = abs_coords
end

function roomResizeUpdate(new_mouse_coords::Tuple{Int,Int})
    border_move_abs = new_mouse_coords .- original_mouse_coords
    border_move_tiles = floor.(Int, border_move_abs ./ 8)
    border_move = border_move_tiles .* 8
    size_change = (isLeftEdge ? -border_move[1] : isRightEdge ? border_move[1] : 0,
                   isTopEdge ? -border_move[2] : isBottomEdge ? border_move[2] : 0)
    # Never limit size for rooms already smaller in one dimension
    if subtool=="Resize"
        desired_size = max.(room_original_size .+ size_change, roomMinSize)
    else # Filler Resize removes size limit
        desired_size = max.(room_original_size .+ size_change, (16, 16))  # 2 by 2 tiles
    end
    actual_size_change = desired_size .- room_original_size
    position_change = (isLeftEdge ? -actual_size_change[1] : 0, isTopEdge ? -actual_size_change[2] : 0)
    desired_position = room_original_position .+ position_change
    if desired_size == relevantRoom.size
        return
    end
    # Update room position and size
    # placements / tiles may be moved out of the room. Placements still exist after resize,
    # but tiles outside the room are cleared when the resize ends
    
    entity_move = relevantRoom.position .- desired_position
    relevantRoom.size = desired_size
    relevantRoom.position = desired_position

    total_position_move = desired_position .- room_original_position
    updateTiles!(total_position_move, Maple.tile_fg_names["Air"], Maple.tile_fg_names["Air"])
    # Entities are moved by only the most recent movement, not the total movement
    if entity_move !== (0, 0)
        updateEntitiesTriggersDecals(entity_move)
    end
    updateRoomList()
    markForRedraw(relevantRoom, Ahorn.loadedState.map)
    Ahorn.draw(Ahorn.canvas)
end

# Adapted from Maple
# Set the fg and bg tiles to the size of the room
function updateTiles!(total_pos_change::Tuple{Int, Int}, center::Char='0', border::Char='0')
    # Room size already changed
    tw, th = ceil.(Int, relevantRoom.size ./ 8)
    dw, dh = floor.(Int, total_pos_change ./ 8)
    # positive = down/right

    # Create new tiles arrays for the room, filled with default (air)
    tiles = fill(center, (th, tw))
    fg = copy(tiles)
    bg = copy(tiles)
    objTiles = fill(-1, (th, tw))

    # Copy the original tiles to the new tiles
    # Size of what will be copied (window size)
    fth, ftw = min.(size(room_original_tiles[1]), (th, tw))
    bth, btw = min.(size(room_original_tiles[2]), (th, tw))
    oth, otw = min.(size(room_original_tiles[3]), (th, tw))

    # Goal is that original tiles move in the room
    # to stay in the same place on the map

    # If position decreased dh (resized up/left -> size increased)
    # offset copy location dh to match original position
    # (result window moved down/right in room, air in the new space at top/left)
    rh = max(0, -dh)
    rw = max(0, -dw)
    # If position increased dh (eg resized down from top)
    # offset source location dh to match original position
    # (source window moved down/right, to align with bottom/right edges)
    sh = max(0, dh)
    sw = max(0, dw)

    fg[1+rh:fth+rh, 1+rw:ftw+rw] = room_original_tiles[1][1+sh:fth+sh, 1+sw:ftw+sw]
    bg[1+rh:bth+rh, 1+rw:btw+rw] = room_original_tiles[2][1+sh:bth+sh, 1+sw:btw+sw]
    objTiles[1+rh:oth+rh, 1+rw:otw+rw] = room_original_tiles[3][1+sh:oth+sh, 1+sw:otw+sw]

    relevantRoom.fgTiles = Tiles(fg)
    relevantRoom.bgTiles = Tiles(bg)
    relevantRoom.objTiles = ObjectTiles(objTiles)
end

#TODO can/should we call Ahorn.moved() on these?
function updateEntitiesTriggersDecals(entity_move::Tuple{Int, Int})
    for entity in relevantRoom.entities
        applyMovementAll!(entity, entity_move)
    end
    for trigger in relevantRoom.triggers
        applyMovementAll!(trigger, entity_move)
    end
    for decal in relevantRoom.fgDecals
        applyMovementAll!(decal, entity_move)
    end
    for decal in relevantRoom.bgDecals
        applyMovementAll!(decal, entity_move)
    end
end

function markForRedraw(room::Maple.Room, map::Maple.Map)
    dr = Ahorn.getDrawableRoom(map, room)
    for layer in dr.layers
        layer.redraw = true
    end
end

applyMovementAll!(target, ox, oy) = false

function applyMovementAll!(target::Union{Maple.Entity, Maple.Trigger}, movement)
    ox, oy = movement
    target.data["x"] += ox
    target.data["y"] += oy
    nodes = get(target.data, "nodes", [])
    for node = 1:size(nodes)[1]
        nodes[node] = nodes[node] .+ movement
    end
    return true
end

function applyMovementAll!(decal::Maple.Decal, movement)
    ox, oy = movement
    decal.x += ox
    decal.y += oy
    return true
end


function roomResizeEnd()
    if resizingRoom
        # The room, tiles, and placements are already in place, we just have to stop resizing.
        global resizingRoom = false
        updateCursor()
        # reset other things just in case
        global room_original_position = nothing
        global room_original_size = nothing
        global room_original_tiles = (nothing, nothing, nothing)
        global original_mouse_coords = nothing
    end
end

end #module
